# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

import contextlib
import io
import threading
from typing import List

from fastapi import FastAPI, HTTPException, Query, Response, status

from huemon.commands.command_handler import create_default_command_handler
from huemon.infrastructure.logger_factory import create_logger
from huemon.utils.plugins import get_command_plugins_path

LOG = create_logger()


class HuemonServerFactory:  # pylint: disable=too-few-public-methods
    __stdout_reader_mutex = threading.Lock()

    @staticmethod
    def __create_command_route_handler(command_handler, command_name: str):
        empty_query = Query([])

        def handle_command_route(
            q: List[str] = empty_query,
        ):  # pylint: disable=invalid-name
            with HuemonServerFactory.__stdout_reader_mutex:
                context_reader = io.StringIO()
                with contextlib.redirect_stdout(context_reader):
                    try:
                        command_handler.exec(command_name, q)
                    except SystemExit as system_exit:
                        raise HTTPException(
                            detail=context_reader.getvalue(),
                            status_code=status.HTTP_400_BAD_REQUEST,
                        ) from system_exit

            return Response(context_reader.getvalue(), media_type="plain/text")

        return handle_command_route

    @staticmethod
    def create(config: dict) -> FastAPI:
        app = FastAPI()

        command_handler = create_default_command_handler(
            config, get_command_plugins_path(config)
        )

        for command_name in command_handler.available_commands():
            app.add_api_route(
                f"/{command_name}",
                HuemonServerFactory.__create_command_route_handler(
                    command_handler, command_name
                ),
            )

        return app



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.


from fcntl import LOCK_EX, LOCK_NB, flock

from huemon.infrastructure.logger_factory import create_logger

LOG = create_logger()


def run_locked(lock_file, fn_call):
    with open(lock_file, "w") as f_lock:
        try:
            flock(f_lock.fileno(), LOCK_EX | LOCK_NB)
            LOG.debug("Acquired lock successfully (file=%s)", lock_file)

            return fn_call()
        except BlockingIOError:
            LOG.debug("Failed to acquire lock (file=%s)", lock_file)
        except Exception as error:  # pylint: disable=broad-except
            LOG.debug(
                "Something unexpected went wrong while acquiring lock (file=%s, error=%s)",
                lock_file,
                error,
            )

    return None



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

import os
from functools import reduce
from typing import Dict, List, Type

from pyella.either import Either, rights
from pyella.maybe import Maybe, maybe, of

from huemon.api.api_interface import ApiInterface
from huemon.commands.hue_command_interface import HueCommand
from huemon.discoveries.discovery_interface import Discovery
from huemon.infrastructure.logger_factory import create_logger
from huemon.infrastructure.plugin_loader import load_plugins
from huemon.utils.assertions import assert_exists, assert_num_args
from huemon.utils.paths import create_local_path
from huemon.utils.plugins import get_discovery_plugins_path

LOG = create_logger()


def create_discovery_handlers(
    api: ApiInterface, plugins: List[Type[Discovery]]
) -> Dict[str, Discovery]:
    return reduce(lambda p, c: {**p, c.name(): c(api)}, plugins, {})


class DiscoveryHandler:  # pylint: disable=too-few-public-methods
    def __init__(self, handlers):
        self.handlers = handlers

    def exec(self, discovery_type):
        LOG.debug(
            "Running `%s` command (discovery_type=%s)",
            DiscoverCommand.name(),
            discovery_type,
        )
        target, maybe_sub_target, *_ = discovery_type.split(":") + [None]

        assert_exists(list(self.handlers), target)

        self.handlers[target].exec([maybe_sub_target] if maybe_sub_target else [])

        LOG.debug(
            "Finished `%s` command (discovery_type=%s)",
            DiscoverCommand.name(),
            discovery_type,
        )


class Discover:  # pylint: disable=too-few-public-methods
    def __init__(self, config: dict, api: ApiInterface):
        self.api = api

        self.discovery_plugins_path = get_discovery_plugins_path(config)

        self.handler = self.__create_discovery_handler()

    def __create_discovery_handler(self):
        LOG.debug("Loading discovery plugins (path=%s)", self.discovery_plugins_path)
        discovery_handler_plugins = create_discovery_handlers(
            self.api,
            rights(
                Discover.__load_plugins_and_hardwired_handlers(
                    of(self.discovery_plugins_path)
                )
            ),
        )
        LOG.debug(
            "Finished loading discovery plugins (path=%s)", self.discovery_plugins_path
        )

        return DiscoveryHandler(discovery_handler_plugins)

    @staticmethod
    def __load_discovery_plugin(path: str):
        return load_plugins("discovery", path, Discovery)

    @staticmethod
    def __load_plugins_and_hardwired_handlers(
        discovery_plugins_path: Maybe[str],
    ) -> List[Either[str, Type[Discovery]]]:
        hardwired_discoveries_path = create_local_path(
            os.path.join("discoveries", "internal")
        )

        return maybe(
            [], Discover.__load_discovery_plugin, discovery_plugins_path
        ) + Discover.__load_discovery_plugin(hardwired_discoveries_path)

    def discover(self, discovery_type):
        self.handler.exec(discovery_type)


class DiscoverCommand(HueCommand):
    def __init__(
        self, config: dict, api: ApiInterface
    ):  # pylint: disable=super-init-not-called
        self.discovery = Discover(config, api)

    @staticmethod
    def name():
        return "discover"

    def exec(self, arguments):
        LOG.debug(
            "Running `%s` command (arguments=%s)", DiscoverCommand.name(), arguments
        )
        assert_num_args(1, arguments, DiscoverCommand.name())

        discovery_type, *_ = arguments

        self.discovery.discover(discovery_type)
        LOG.debug("Finished `discover` command (arguments=%s)", arguments)



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

from huemon.infrastructure.logger_factory import create_logger
from huemon.infrastructure.urllib_safe_opener import (
    security_urllib_allow_http_and_https_schemas_only,
)

LOG = create_logger()


def bootstrap():
    LOG.debug("Bootstrapping application")
    security_urllib_allow_http_and_https_schemas_only()
    LOG.debug("Finished bootstrapping application")



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.


import json
import os
import tempfile


def cache_output_to_temp(cache_file_path, fn_call):
    tmp_fd, tmp_file_path = tempfile.mkstemp()
    with open(tmp_file_path, "w") as f_tmp:
        f_tmp.write(json.dumps(fn_call()))

    os.close(tmp_fd)

    os.rename(tmp_file_path, cache_file_path)

    with open(cache_file_path) as f_json:
        return json.loads(f_json.read())



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
from pathlib import Path


def get_root_module_path():
    return Path(sys.modules["__main__"].__file__).parent


def create_local_path(relative_path: str):
    return str(os.path.join(get_root_module_path().absolute(), relative_path))



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

import sys
from typing import NamedTuple

from huemon.infrastructure.logger_factory import create_logger
from huemon.utils.const import EXIT_FAIL

LOG = create_logger()

E_CODE_PLUGIN_LOADER = -2


class HueError(NamedTuple):
    code: int
    message: str = ""
    context: dict = {}


def exit_fail(message, *arguments):
    LOG.error(message, *arguments)
    print(message % tuple(arguments))

    sys.exit(EXIT_FAIL)



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.


def __create_plugins_path(plugin_type: str, config: dict, fallback_path: str = None):
    plugins_section_exists = "plugins" in config
    if not plugins_section_exists:
        return fallback_path

    plugin_type_exists = config["plugins"][plugin_type]
    if not plugin_type_exists:
        return fallback_path

    return (
        config["plugins"][plugin_type]["path"]
        if "path" in config["plugins"][plugin_type]
        else fallback_path
    )


def get_command_plugins_path(config: dict, fallback_path: str = None):
    return __create_plugins_path("commands", config, fallback_path)


def get_discovery_plugins_path(config: dict, fallback_path: str = None):
    return __create_plugins_path("discoveries", config, fallback_path)



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

import uvicorn  # type: ignore

from huemon.api.api_interface import ApiInterface
from huemon.api_server import HuemonServerFactory
from huemon.commands.hue_command_interface import HueCommand
from huemon.infrastructure.logger_factory import create_logger
from huemon.utils.assertions import assert_exists, assert_num_args

LOG = create_logger()


class MyServer:  # pylint: disable=too-few-public-methods
    @staticmethod
    def start(config: dict):
        has_server_config = "server" in config

        host = (
            config["server"]["host"]
            if has_server_config and "host" in config["server"]
            else "127.0.0.1"
        )
        port = (
            int(config["server"]["port"])
            if has_server_config and "port" in config["server"]
            else 8000
        )

        uvicorn.run(HuemonServerFactory.create(config), host=host, port=port)


class AgentCommand(HueCommand):
    __SYSTEM_ACTION_MAP = {
        "start": MyServer.start,
    }

    def __init__(
        self, config: dict, _: ApiInterface
    ):  # pylint: disable=unused-argument
        self.config = config

    @staticmethod
    def name():
        return "agent"

    def exec(self, arguments):
        LOG.debug("Running `%s` command (arguments=%s)", AgentCommand.name(), arguments)
        assert_num_args(1, arguments, AgentCommand.name())

        action, *_ = arguments

        assert_exists(list(AgentCommand.__SYSTEM_ACTION_MAP), action)

        HueCommand._process(self.__SYSTEM_ACTION_MAP[action](self.config))
        LOG.debug(
            "Finished `%s` command (arguments=%s)", AgentCommand.name(), arguments
        )



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

from huemon.utils.errors import exit_fail


def assert_num_args(expected_number_of_arguments: int, arguments: list, context: str):
    if len(arguments) != expected_number_of_arguments:
        argument_text = "argument" if expected_number_of_arguments == 1 else "arguments"

        exit_fail(
            "Expected exactly %s %s for `%s`, received %s",
            expected_number_of_arguments,
            argument_text,
            context,
            len(arguments),
        )


def assert_exists(expected_values: list, value: str):
    if value not in expected_values:
        exit_fail("Received unknown value `%s` (expected=%s)", value, expected_values)



# Copyright (c) Ely Deckers.
#
# This source code is licensed under the MPL-2.0 license found in the
# LICENSE file in the root directory of this source tree.

import importlib.util
import inspect
from pathlib import Path
from typing import List, Tuple, Type, TypeVar, cast

from pyella.either import Either, Left, right
from pyella.maybe import Maybe

from huemon.utils.errors import E_CODE_PLUGIN_LOADER, HueError

TA = TypeVar("TA")


def __error(message: str) -> HueError:
    return HueError(E_CODE_PLUGIN_LOADER, message)


def __lerror(message: str) -> Left[HueError, Type[TA]]:
    return Left(__error(message))


def __get_plugin_type(
    module_name: str, path: str, sub_class: Type[TA]
) -> Either[HueError, Type[TA]]:
    return (
        Maybe.of(importlib.util.spec_from_file_location(module_name, path))
        .maybe(
            __lerror("ModuleSpec could not be loaded from the provided path"),
            right,
        )
        .bind(
            lambda spec: Maybe.of(spec.loader).maybe(
                __lerror("ModuleSpec has no loader"), lambda _: right(spec)
            )
        )
        .bind(
            lambda spec: Maybe.of(importlib.util.module_from_spec(spec)).maybe(
                __lerror("Module could not be loaded from ModuleSpec"),
                lambda module: right((spec, module)),
            ),
        )
        .discard(
            lambda sm: Either.pure(
                Maybe.of(sm[0].loader).bind(
                    lambda loader: Maybe.of(loader.exec_module(sm[1]))
                )
            )
        )
        .fmap(lambda sm: sm[1])
        .fmap(
            lambda module: list(
                map(
                    lambda obj: cast(Tuple[str, Type[TA]], obj),
                    filter(
                        lambda member: inspect.isclass(member[1])
                        and issubclass(member[1], sub_class)
                        and member[1] is not sub_class,
                        inspect.getmembers(module),
                    ),
                ),
            ),
        )
        .bind(
            lambda plugin_types: Maybe.of(len(plugin_types) > 0).maybe(
                __lerror(f"No plugin of type `{sub_class}` found"),
                lambda _: right(plugin_types[0][1]),
            )
        )
    )


def load_plugins(
    module_name: str, path: str, plugin_type: Type[TA]
) -> List[Either[HueError, Type[TA]]]:
    return list(
        map(
            lambda plugin_path: __get_plugin_type(
                f"{module_name}.{plugin_path.stem}",
                str(plugin_path.absolute()),
                plugin_type,
            ),
            Path(path).glob("*.py"),
        )
    )
